# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os, sys 
import pickle 
import multiprocessing as mp 
import collections
  

class Tagging(object):
    
    def __init__(self):
        # get args
        self.resource_path = sys.argv[1]
        self.workspace =  sys.argv[2]
        self.corpus_name =  sys.argv[3]
        self.language =  sys.argv[4]
        self.source_folder =  sys.argv[5]
        # load sent tokenizer
        sent_tokenizer_path = self.resource_path + 'sent_tokenizer_' + self.language + '.pickle'
        with open(sent_tokenizer_path, 'rb') as fh:
            self.sent_tokenizer = pickle.load(fh)
        # load word tokenizer
        word_tokenizer_path = self.resource_path + 'word_tokenizer_' + self.language +  '.pickle'
        with open(word_tokenizer_path, 'rb') as fh:
            self.word_tokenizer = pickle.load(fh)
        # load tagger 
        tagger_path = self.resource_path+ 'pos_tagger_' + self.language +  '.pickle'
        with open(tagger_path, 'rb') as fh:
            self.tagger = pickle.load(fh)
        # normalize source_folder path
        if not self.source_folder.endswith('/'):
            self.source_folder = self.source_folder + '/'
        
    def get_files(self):
        i = 0
        for filename in os.listdir(self.source_folder):
            i+=1
            yield (filename,i)
    
    def get_wfiles(self):
        i = 0
        for filename in os.listdir(self.workspace+self.corpus_name + '/data/tmp2/'):
            i+=1
            yield filename
    
    def get_tfiles(self):
        i = 0
        for filename in os.listdir(self.workspace+self.corpus_name + '/data/tmp3/'):
            i+=1
            yield filename 
    
    def sent2str(self,tagged_sent,sent_id,file_id):
        s = []
        for token in tagged_sent:
            if token[0] not in self.unique_w:
                self.unique_w[token[0]]=None
            if token[1] not in self.unique_t:
                self.unique_t[token[1]]=None
            s.append("%s\t%s\t%s\t%s" % (token[0],token[1],sent_id,file_id))
        return '\n'.join(s)
            
    
    def tag_file(self,filename):
        self.unique_w = collections.defaultdict()
        self.unique_t = collections.defaultdict()
        sent_id = 0
        tagged_sents = []
        with open(self.source_folder + filename[0],'r',encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip()) != 0:
                    sents = self.sent_tokenizer.tokenize(line.strip())
                    for sent in sents:
                        sent_id+=1
                        tagged_sents.append(self.sent2str(self.tagger.tag(self.word_tokenizer.tokenize(sent)),sent_id,filename[1]))
        
        with open(self.workspace + self.corpus_name + '/data/tmp1/' + filename[0],'w',encoding='utf-8') as fh2:
            fh2.write('\n'.join(tagged_sents))

        with open(self.workspace+self.corpus_name + '/data/tmp2/' + filename[0],'wb') as fh:
            pickle.dump(self.unique_w,fh)
        
        with open(self.workspace+self.corpus_name + '/data/tmp3/' + filename[0],'wb') as fh:
            pickle.dump(self.unique_t,fh)
    
    def unique(self):
        idx = collections.defaultdict()
        i = 0
        # word indexes
        files = self.get_wfiles()
        for filename in files:
            with open(self.workspace + self.corpus_name + '/data/tmp2/' + filename,'rb') as fh:
                d = pickle.load(fh)
            for k in d:
                if k not in idx:
                    i+=1
                    idx[k]=i
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','wb' ) as fh:
            pickle.dump(idx,fh) 
        # tag indexes
        idx = collections.defaultdict()
        i = 0
        files = self.get_tfiles()
        for filename in files:
            with open(self.workspace + self.corpus_name + '/data/tmp3/' + filename,'rb') as fh:
                d = pickle.load(fh)
            for k in d:
                if k not in idx:
                    i+=1
                    idx[k]=i
        with open(self.workspace + self.corpus_name + '/data/idx/tags.pickle','wb' ) as fh:
            pickle.dump(idx,fh)
        # file indexes
        idx = collections.defaultdict()
        i = 0
        files = self.get_files()
        for filename in files:
            i+=1
            idx[filename]=i 
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','wb' ) as fh:
            pickle.dump(idx,fh)
    
    
    def execute(self):
        files = self.get_files()
        p = mp.Pool(mp.cpu_count())
        p.map(self.tag_file,files)
        p.close()
        p.join()
        self.unique()
          

if __name__ == '__main__':
    t = Tagging()
    t.execute()
    
    
    


# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os 
import numpy as np 
import pickle


class KitData(object):
    
    def __init__(self,workspace,corpus_name,language):
        self.workspace = workspace 
        self.corpus_name = corpus_name 
        self.language = language  

    def info_get(self):
        if os.path.exists(self.workspace + self.corpus_name + '/data/idx/info.pickle'):
            with open (self.workspace + self.corpus_name + '/data/idx/info.pickle', 'rb') as fh:
                info = pickle.load(fh)
            return info
        else:
            return None
        
    def tokens_get(self):
        info = self.info_get()
        return info[0]
    
    def types_get(self):
        info = self.info_get()
        return info[1]
    
    def ttr_get(self):
        info = self.info_get()
        return info[2]
    
    def hapax_get(self):
        info = self.info_get()
        return info[3]
    
    def textfiles_get_names(self):
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            d = pickle.load(fh)
        for k in d:
            yield k[0]
    
    def fileids_get(self):
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            d = pickle.load(fh)
        for k in d:
            yield k[1]
    
    def words_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','rb') as fh:
            dict_words = pickle.load(fh)
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                for i in range(0,arr.shape[0]):
                    yield dict_words[arr[i,0]]
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                    for i in range(0,arr.shape[0]):
                        yield dict_words[arr[i,0]]
    
    def tagged_words_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','rb') as fh:
            dict_words = pickle.load(fh)
        with open(self.workspace + self.corpus_name + '/data/idx/tags.pickle','rb') as fh:
            dict_tags = pickle.load(fh)
            
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                for i in range(0,arr.shape[0]):
                    yield (dict_words[arr[i,0]],dict_tags[arr[i,1]])
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                    for i in range(0,arr.shape[0]):
                        yield (dict_words[arr[i,0]],dict_tags[arr[i,1]])
    
    def sents_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','rb') as fh:
            dict_words = pickle.load(fh)
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                arr = np.split(arr[:, 0], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                for a in arr:
                    sent = []
                    for i in a:
                        sent.append(dict_words[i])
                    yield tuple(sent)
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                    arr = np.split(arr[:, 0], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                    for a in arr:
                        sent = []
                        for i in a:
                            sent.append(dict_words[i])
                        yield tuple(sent)
    
    def tagged_sents_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','rb') as fh:
            dict_words = pickle.load(fh)
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        with open(self.workspace + self.corpus_name + '/data/idx/tags.pickle','rb') as fh:
            dict_tags = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                arr_t = np.split(arr[:, 1], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                arr = np.split(arr[:, 0], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                for j in range(0,len(arr)):
                    sent = []
                    for i in range(0,arr[j].shape[0]):
                        sent.append((dict_words[arr[j][i]],dict_tags[arr_t[j][i]]))
                    yield sent 
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                    arr_t = np.split(arr[:, 1], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                    arr = np.split(arr[:, 0], np.cumsum(np.unique(arr[:, 2], return_counts=True)[1])[:-1])
                    for j in range(0,len(arr)):
                        sent = []
                        for i in range(0,arr[j].shape[0]):
                            sent.append((dict_words[arr[j][i]],dict_tags[arr_t[j][i]]))
                        yield sent
    
    def ndarrays_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                yield arr 
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    arr = np.load(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
                    yield arr
    
    def ndarrays_filenames_get(self,fileids=None):
        with open(self.workspace + self.corpus_name + '/data/idx/filenames.pickle','rb') as fh:
            dict_filenames = pickle.load(fh)
        if fileids == None:
            for k in dict_filenames:
                yield str(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
        else:
            if type(fileids) != list:
                fileids = [fileids] 
            for k in dict_filenames:
                if k[1] in fileids:
                    yield str(self.workspace + self.corpus_name + '/data/npy/' + k[0] + '.npy')
    
    def dict_words_get(self):
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','rb') as fh:
            return pickle.load(fh)
    
    def dict_tags_get(self):
        with open(self.workspace + self.corpus_name + '/data/idx/tags.pickle','rb') as fh:
            return pickle.load(fh)
        
    
    
                    
    
                
        
                            
                
        
        
    
    
    
        
    
     
        


# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os, sys,subprocess  
from cmd import Cmd
import argparse
from kitconc import version 
from kitconc.kit_corpus import Corpus
from kitconc.kit_plots import CollGraph
from kitconc import kit_util  
from kitconc.core import Examples 
import re 
import pandas as pd 


  
class Kit(Cmd):
    
    def parse_arg(self,arg):
        if len(arg.strip()) != 0:
            arg = arg.replace('\t','')
            multi_spaces = re.compile(' +')
            multi_tabs = re.compile('\t+')
            arg = multi_spaces.sub(' ',arg)
            arg = arg.strip()
            args = []
            if arg.find('"') != -1 or arg.find("'") != -1:
                arg = arg.replace("'",'"')
                s = []
                status = 0
                for c in arg:
                    if status == 0 and c == '"':
                        status = 1
                    elif status == 0 and c == ' ':
                        s.append('\t')
                    elif status == 1 and c == '"':
                        s.append('\t')
                        status = 0
                    else:
                        s.append(c)
                args = ''.join(s)
                args = multi_tabs.sub('\t',args)
                args = args.split('\t')
            else:
                args = arg.split(' ')
        else:
            args = [] 
        
        if type(args) == list:
            arr = []
            if len(args) > 1:
                for a in args:
                    if len(a)!=0:
                        arr.append(a)
                args = arr 
        return args
    
    def get_parser(self,function_name):
        parser = None
        if function_name == 'help':
            parser = argparse.ArgumentParser()
            parser.prog= "help"
            parser.description='Shows help.'
        
        elif function_name == 'workspace':
            parser = argparse.ArgumentParser()
            parser.prog= "workspace"
            parser.description='Sets or shows the current workspace.'
            parser.add_argument('path', type=str, nargs='+',help='folder path for workspace')
        
        elif function_name == 'ls':
            parser = argparse.ArgumentParser()
            parser.prog= "ls"
            parser.description='Lists corpora or files in the current workspace or output folder.'
        
        elif function_name == 'exit':
            parser = argparse.ArgumentParser()
            parser.prog= "exit"
            parser.description='Ends the program.'
        
        elif function_name == 'quit':
            parser = argparse.ArgumentParser()
            parser.prog= "quit"
            parser.description='Ends the program.'
        
        elif function_name == 'version':
            parser = argparse.ArgumentParser()
            parser.prog= "version"
            parser.description='Show version.'
        
        elif function_name == 'cls':
            parser = argparse.ArgumentParser()
            parser.prog= "cls"
            parser.description='Clears screen.'
        
        elif function_name == 'clear':
            parser = argparse.ArgumentParser()
            parser.prog= "clear"
            parser.description='Clears screen.'
        
        elif function_name == 'home':
            parser = argparse.ArgumentParser()
            parser.prog= "home"
            parser.description='Returns to the current workspace root without any corpus selection.'
        
        elif function_name == 'delete':
            parser = argparse.ArgumentParser()
            parser.prog= "delete"
            parser.description='Deletes a corpus.'
        
        elif function_name == 'cleanse':
            parser = argparse.ArgumentParser()
            parser.prog= "cleanse"
            parser.description='Deletes all files from the output folder.'
            
        elif function_name == 'create':
            parser = argparse.ArgumentParser()
            parser.prog= "create"
            parser.description='Creates a new corpus from text files in a source folder.'
            parser.add_argument('name', type=str, nargs='?',help='corpus name for identification')
            parser.add_argument('source', type=str, nargs='?',help='source folder to add textfiles to the corpus')
            parser.add_argument('language', type=str, nargs='?',help='language')
            
        elif function_name == 'use':
            parser = argparse.ArgumentParser()
            parser.prog= "use"
            parser.description='Sets the current corpus in use.'
            parser.add_argument('name', type=str, nargs='?',help='corpus name')
        
        elif function_name =='open':
            parser = argparse.ArgumentParser()
            parser.prog= "open"
            parser.description='Opens a .xlsx file in the output folder using the default program.'
            parser.add_argument('filename', type=str, nargs='?',help='the name of the .xlsx file')
            
        elif function_name == 'wordlist':
            parser = argparse.ArgumentParser()
            parser.prog= "wordlist"
            parser.description='Makes a frequency word list from the current corpus'
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
        
        elif function_name == 'keywords':
            parser = argparse.ArgumentParser()
            parser.prog= "keywords"
            parser.description='Extracts keywords from the current corpus.'
            parser.add_argument('--measure', action='store', dest='measure', type=str, help='statistic measure (chi-square or log-likelihood) default: log-likelihood')
            parser.add_argument('--stoplist', action='store', dest='stoplist', type=str, help='use the stoplist to filter keywords: (y or n) default:n')
        
        elif function_name == 'kwic':
            parser = argparse.ArgumentParser()
            parser.prog= "kwic"
            parser.description='Generates concordance lines in KWIC format.'
            parser.add_argument('node', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word')
            parser.add_argument('--case', action='store', dest='case', type=str, help='Case sensitive: (y or n) default: y')
            parser.add_argument('--regexp', action='store', dest='regexp', type=str, help='Enable case sensitive and regular expression for search word matching (y or n) default: n')
            parser.add_argument('--horizon', action='store', dest='horizon', type=int, help='Left and right horizon of words - default: 7')
            parser.add_argument('--width', action='store', dest='width', type=int, help='horizon width in characters')
            parser.add_argument('--sort1', action='store', dest='sort1', type=str, help='first sorting (L1,L2,L3,L4,L5,R1,R2,R3,R4 or R5)')
            parser.add_argument('--sort2', action='store', dest='sort2', type=str, help='second sorting (L1,L2,L3,L4,L5,R1,R2,R3,R4 or R5)')
            parser.add_argument('--sort3', action='store', dest='sort3', type=str, help='third sorting (L1,L2,L3,L4,L5,R1,R2,R3,R4 or R5)')
            parser.add_argument('--highlight', action='store', dest='highlight', type=str, help='highlights horizon (Example: L1 R1)')
            parser.add_argument('--limit', action='store', dest='limit', type=int, help='Number of concordance lines to return')
        
        elif function_name == 'concordance':
            parser = argparse.ArgumentParser()
            parser.prog= "concordance"
            parser.description='Generates concordance lines.'
            parser.add_argument('node', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word')
            parser.add_argument('--case', action='store', dest='case', type=str, help='Case sensitive: (y or n) default: y')
            parser.add_argument('--regexp', action='store', dest='regexp', type=str, help='Enable case sensitive and regular expression for search word matching (y or n) default: n')
            parser.add_argument('--limit', action='store', dest='limit', type=int, help='Number of concordance lines to return')
        
        elif function_name == 'collocates':
            parser = argparse.ArgumentParser()
            parser.prog= "collocates"
            parser.description='Extracts a list of collocates.'
            parser.add_argument('node', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word')
            parser.add_argument('--coll_pos', action='store', dest='coll_pos', type=str, help='Part of speech (POS) for collocates')
            parser.add_argument('--case', action='store', dest='case', type=str, help='Case sensitive: (y or n) default: y')
            parser.add_argument('--regexp', action='store', dest='regexp', type=str, help='Enable case sensitive and regular expression for search word matching (y or n) default: n')
            parser.add_argument('--left_span', action='store', dest='left_span', type=int, help='left span size for collocates')
            parser.add_argument('--right_span', action='store', dest='right_span', type=int, help='right span size for collocates')
        
        elif function_name == 'collgraph':
            parser = argparse.ArgumentParser()
            parser.prog= "collgraph"
            parser.description='Extracts a list of collocates and shows a plot.'
            parser.add_argument('node', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word')
            parser.add_argument('--coll_pos', action='store', dest='coll_pos', type=str, help='Part of speech (POS) for collocates')
            parser.add_argument('--case', action='store', dest='case', type=str, help='Case sensitive: (y or n) default: y')
            parser.add_argument('--regexp', action='store', dest='regexp', type=str, help='Enable case sensitive and regular expression for search word matching (y or n) default: n')
            parser.add_argument('--left_span', action='store', dest='left_span', type=int, help='left span size for collocates')
            parser.add_argument('--right_span', action='store', dest='right_span', type=int, help='right span size for collocates')
        
        elif function_name == 'wtfreq':
            parser = argparse.ArgumentParser()
            parser.prog= "wtfreq"
            parser.description='Generates a word tag frequency list based on corpus text files.'
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
        
        elif function_name == 'wfreqinfiles':
            parser = argparse.ArgumentParser()
            parser.prog= "wfreqinfiles"
            parser.description='Generates a frequency list based on word occurrence in corpus text files.'
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
        
        elif function_name == 'clusters':
            parser = argparse.ArgumentParser()
            parser.prog= "clusters"
            parser.description='Generates a list of clusters.'
            parser.add_argument('word', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word ')
            parser.add_argument('--size', action='store', dest='size', type=int, help='size of the cluster')
            parser.add_argument('--minfreq', action='store', dest='minfreq', type=int, help='minimum frequency of the cluster')
            parser.add_argument('--minrange', action='store', dest='minrange', type=int, help='minimum range frequency')
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
            
        elif function_name == 'ngrams':
            parser = argparse.ArgumentParser()
            parser.prog= "ngrams"
            parser.description='Generates a list of n-grams.'
            parser.add_argument('size', type=int, nargs='?',help='size of the cluster (default: 3)')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each word ')
            parser.add_argument('--minfreq', action='store', dest='minfreq', type=int, help='minimum frequency of the cluster')
            parser.add_argument('--minrange', action='store', dest='minrange', type=int, help='minimum range frequency')
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
        
        elif function_name == 'dispersion':
            parser = argparse.ArgumentParser()
            parser.prog= "dispersion"
            parser.description='Generates dispersion plots.'
            parser.add_argument('node', type=str, nargs='?',help='search word or phrase')
            parser.add_argument('--pos', action='store', dest='pos', type=str, help='Part of speech (POS) for each search word')
            parser.add_argument('--case', action='store', dest='case', type=str, help='Case sensitive: (y or n) default: y')
            parser.add_argument('--regexp', action='store', dest='regexp', type=str, help='Enable case sensitive and regular expression for search word matching (y or n) default: n')
            parser.add_argument('--limit', action='store', dest='limit', type=int, help='Number of concordance lines to return')
        
        elif function_name == 'keywords_dispersion':
            parser = argparse.ArgumentParser()
            parser.prog= "keywords_dispersion"
            parser.description='Generates dispersion plots from keywords.'
            parser.add_argument('--lowercase', action='store', dest='lowercase', type=str, help='y or n default: y')
            parser.add_argument('--limit', action='store', dest='limit', type=int, help='Maximum number of keywords to use default: 25')
        
        elif function_name == 'text2utf8':
            parser = argparse.ArgumentParser()
            parser.prog= "text2utf8"
            parser.description='Converts textfiles to utf-8 encoding in a source folder.'
            parser.add_argument('source', type=str, nargs='?',help='source folder with textfiles to be converted')
            parser.add_argument('target', type=str, nargs='?',help='target folder to save the converted files')
        
        elif function_name == 'examples':
            parser = argparse.ArgumentParser()
            parser.prog= "examples"
            parser.description='Downloads examples from github.'
            parser.add_argument('--dest_path', action='store', dest='dest_path', type=str, help='Path to store examples')
            
            
        
        
        else:
            parser = None
        return parser
    
    #
    # OVERRIDES HELP 
    #
    
    def do_help(self,arg):
        if len(arg.strip())!= 0:
            parser = self.get_parser(arg)
            if parser != None:
                print('')
                parser.print_help() 
                print('')
        else:
            commands = []
            commands.append('\nDocumented commands (type help <topic>):')
            commands.append('========================================')
            function_names = []
            for item in sorted([func for func in dir(self) if callable(getattr(self, func))]):
                if item.startswith('do_'):
                    function_names.append( item.replace('do_',''))
            i = 0
            lines = []
            for func in function_names:
                i+=1
                lines.append( str(i) + ') ' + func)
            commands.append('\n'.join(lines))
            print('\n'.join(commands))
            print('')
            
    
    # load workspace
    def load_workspace(self):
        """Loads workspace """
        self.workspace = None
        self.corpus_in_use = None
        try:
            if os.path.exists('kitconc.tmp'):
                with open('kitconc.tmp','r') as fh:
                    workspace = str(fh.read()).strip()
                    if os.path.exists(workspace) and os.path.isdir(workspace):
                        self.workspace = workspace.strip()
        except Exception as e:
                print(e)
    
    def do_exit(self, args):
        """\nEnds program\n"""
        print ("Finishing...")
        raise SystemExit
    
    def do_quit(self,args):
        """\nEnds program\n"""
        print ("Finishing...")
        raise SystemExit

    def do_version(self,args):
        """\nShows version\n"""
        print ("\nKitconc " + version.__version__)
        
    
    def do_cls(self,args):
        """\nClear screen\n"""
        os.system('cls' if os.name=='nt' else 'clear') 
    
    def do_clear(self,args):
        """\nClear screen\n"""
        os.system('cls' if os.name=='nt' else 'clear')
    
    def do_ls(self,args):
        """\nDescription: Lists corpora or files in the output corpus folder.
        \nUsage: ls  
        """
        try:
            if self.workspace is not None:
                if self.corpus_in_use == None:
                    files = os.listdir(self.workspace)
                    corpora = []
                    df = pd.DataFrame(columns=['Name','Language','Texts','Tokens','Types','TTR'])
                    
                    for filename in files:
                        if os.path.isdir(self.workspace + filename):
                            if os.path.exists(self.workspace + filename + '/info.tab') == True:
                                tb = pd.read_table(self.workspace + filename + '/info.tab',header=None)
                                reg = []
                                for row in tb.itertuples(index=False):
                                    if row[0] not in ['workspace:','encoding:']:
                                        reg.append(row[1])
                                if len(reg) >= 6:
                                    df = df.append(pd.Series(reg, index=df.columns ), ignore_index=True)
                                corpora.append(filename)
                    if len(df) !=0:
                        print('')
                        print(df.head(len(df)))
                        print('')

                else:
                    output_path = self.workspace + self.corpus_in_use + '/output/'
                    files = os.listdir(output_path)
                    contents = []
                    for filename in files:
                        if os.path.isfile(output_path + filename):
                            contents.append(filename)
                    if len(contents) !=0:
                        print ('\n' +'\n'.join(contents) + '\n')
            else:
                print('\nNo workspace set.\n')
        except Exception as e:
                print(e)
    
    def do_home(self,args):
        """\nDescription: Returns to the current workspace root without any corpus selection.
        \nUsage: home
        """
        try:
            if self.corpus_in_use != None:
                self.corpus_in_use = None
                self.prompt = 'kitconc>'
                print ('')
        except Exception as e:
                print(e)
    
    def do_delete(self,s):
        """\nDescription: Deletes a corpus.
        \nUsage: delete [name]
        """
        try:
            corpus_name = s.strip()
            if self.__corpus_exists(corpus_name):
                flag = input('Do you really want to delete the corpus "' + corpus_name + '"? [y,n]')
                if flag.strip().lower() == 'y' and len(corpus_name) != 0 :
                    import shutil
                    shutil.rmtree(self.workspace + corpus_name)
                    print('')
        except Exception as e:
                print(e)
    
    def do_cleanse(self,s):
        """\nDescription: Deletes all files from the output folder.
        \nUsage: cleanse
        """
        try:
            if self.corpus_in_use != None:
                flag = input('Do you really want to delete all files? [y,n]')
                if flag.strip().lower() == 'y':
                    output_path = self.workspace + self.corpus_in_use + '/output/'
                    files = os.listdir(output_path)
                    if len(files) != 0:
                        for filename in files:
                            os.remove(output_path + filename)
                            print('removing ... ' + filename)
        except Exception as e:
                print(e)
    
    def do_open(self,arg):
        try:
            if self.corpus_in_use != None:
                output_path = self.workspace + self.corpus_in_use + '/output/'
                parser = self.get_parser('open')
                args = parser.parse_args(self.parse_arg(arg))
                if args.filename != None:
                    if args.filename.endswith('.xlsx')==False:
                        args.filename= args.filename + '.xlsx'
                    if os.path.exists(output_path + args.filename):
                        if os.name == 'nt':
                            #on Windows:
                            os.system(output_path + args.filename)
                        else:
                            #on linux:
                            opener ="open" if sys.platform == "darwin" else "xdg-open"
                            output = subprocess.call([opener, output_path + '/' + args.filename])
                            output.trim()
                            output = None
                            print('')
                    else:
                        print('\nFile does not exist.')
                        print('')
        except Exception as e:
                print(e)
    
        
    #
    # WORKSPACE 
    #
    
    def do_workspace(self,args):
        if args is None or len(args.strip())== 0:
            if self.workspace == None:
                print('\nThere is no current workspace.\n')
            else:
                print('\n' + self.workspace + '\n')
        else:
            if os.path.exists(args.strip()) == True and os.path.isdir(args.strip())==True:
                if args.strip().endswith('/'):
                    self.workspace = args 
                else:
                    self.workspace = args + '/'
                with open('kitconc.tmp','w') as fh:
                    fh.write(self.workspace)
            else:
                print('\nNot a valid workspace.\nThe workspace command takes a path as argument.\n')
    
    
    def __corpus_exists(self,corpus_name):
        flag = False
        if self.workspace is not None:
            if os.path.exists(self.workspace + '/' + corpus_name + '/info.tab'):
                flag = True
        return flag 
    
    
    def __corpus_info (self,corpus_name):
        info = {}
        with open(self.workspace + corpus_name + '/info.tab','r',encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip()) !=0:
                    fields = line.strip().split('\t')
                    if len(fields) >= 2:
                        info[fields[0].replace(':','')]=fields[1]
        return info
    
    def do_create(self,arg):
        parser = self.get_parser('create')
        args = parser.parse_args(self.parse_arg(arg))
        corpus = Corpus(self.workspace,args.name,args.language)
        corpus.add_texts(args.source, show_progress=True)
    
    def do_use(self,s):
        """\nDescription: Sets the current corpus in use.
        \nUsage: use [name]
        """
        try:
            corpus_name = s.strip()
            if self.__corpus_exists(corpus_name):
                corpus_name = s.strip()
                self.corpus_in_use = corpus_name
                print('') 
                self.prompt = 'kitconc/' + self.corpus_in_use + '>'
            else:
                print('\nNo corpus found.\n')
        except Exception as e:
                print(e)
    
    def do_wordlist(self,arg):
        parser = self.get_parser('wordlist')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                lowercase = True
                if args.lowercase is not None:
                    if args.lowercase == 'n':
                        lowercase = False 
                wordlist = corpus.wordlist(lowercase=lowercase,show_progress=True)
                wordlist.save_excel(self.workspace + self.corpus_in_use + '/output/wordlist.xlsx')
                wordlist = None
                print('')
        except Exception as e:
                print(e)
    
    def do_keywords(self,arg):
        parser = self.get_parser('keywords')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                measure = corpus.MEASURE_LOGLIKELIHOOD
                stoplist = []
                if args.measure !=None:
                    if args.measure == 'chi-square':
                        measure = corpus.MEASURE_CHI_SQUARE
                if args.stoplist !=None:
                    if args.stoplist == 'y':
                        with open (corpus.resource_data_path + 'stoplist_' + corpus.language + '.tab','r') as fh:
                            for line in fh:
                                if len(line.strip())!= 0:
                                    stoplist.append(line.strip())
                print('Wordlist:')
                wordlist = corpus.wordlist(show_progress=True)
                print('Keywords:')
                keywords = corpus.keywords(wordlist,measure=measure,stoplist=stoplist,show_progress=True)
                keywords.save_excel(self.workspace + self.corpus_in_use + '/output/keywords.xlsx')
                del wordlist
                del keywords 
                print('')
        except Exception as e:
            print(e)
    
    def do_kwic(self,arg):
        parser = self.get_parser('kwic')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_case_sensitive = False
                arg_regexp = False
                arg_horizon = 7
                arg_limit = None
                arg_width = 50
                arg_sort1 = None
                arg_sort2 = None
                arg_sort3 = None
                arg_highlight = None
                if args.pos != None:
                    arg_pos = args.pos 
                if args.case != None:
                    if args.case == 'y':
                        arg_case_sensitive = True
                    else: 
                        arg_case_sensitive = False
                if args.regexp != None:
                    if args.regexp == 'y':
                        arg_regexp = True
                    else:
                        arg_regexp = False
                if args.horizon !=None:
                    arg_horizon = int(args.horizon)  
                if args.limit != None:
                    arg_limit = int(args.limit)
                if args.width != None:
                    arg_width = args.width
                if args.sort1 != None:
                    arg_sort1 = args.sort1
                if args.sort2 != None:
                    arg_sort2 = args.sort2 
                if args.sort3 != None:
                    arg_sort3 = args.sort3 
                if args.highlight != None:
                    arg_highlight = args.highlight
                kwic = corpus.kwic(args.node,pos=arg_pos,case_sensitive=arg_case_sensitive,regexp=arg_regexp,horizon=arg_horizon,limit=arg_limit,show_progress=True)
                if arg_sort1 != None:
                    kwic.sort(arg_sort1, arg_sort2, arg_sort3) 
                kwic.save_excel(self.workspace + self.corpus_in_use + '/output/kwic.xlsx',width=arg_width,highlight=arg_highlight)
                print('')
        except Exception as e:
            print(e)
    
    def do_concordance(self,arg):
        parser = self.get_parser('concordance')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_case_sensitive = False
                arg_regexp = False
                arg_limit = None
                if args.pos != None:
                    arg_pos = args.pos 
                if args.case != None:
                    if args.case == 'y':
                        arg_case_sensitive = True
                    else: 
                        arg_case_sensitive = False
                if args.regexp != None:
                    if args.regexp == 'y':
                        arg_regexp = True
                    else:
                        arg_regexp = False
                if args.limit != None:
                    arg_limit = int(args.limit)
                concordance = corpus.concordance(args.node,pos=arg_pos,case_sensitive=arg_case_sensitive,regexp=arg_regexp,limit=arg_limit,show_progress=True)
                concordance.save_excel(self.workspace + self.corpus_in_use + '/output/concordance.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_collocates(self,arg):
        parser = self.get_parser('collocates')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_coll_pos = None 
                arg_case_sensitive = False
                arg_regexp = False
                arg_left_span = 5
                arg_right_span = 5
                if args.pos != None:
                    arg_pos = args.pos 
                if args.coll_pos !=None:
                    arg_coll_pos = args.coll_pos
                if args.case != None:
                    if args.case == 'y':
                        arg_case_sensitive = True
                    else: 
                        arg_case_sensitive = False
                if args.regexp != None:
                    if args.regexp == 'y':
                        arg_regexp = True
                    else:
                        arg_regexp = False
                if args.left_span != None:
                    arg_left_span = args.left_span
                if args.right_span != None:
                    arg_right_span = args.right_span 
                collocates = corpus.collocates(args.node,pos=arg_pos,coll_pos=arg_coll_pos,left_span=arg_left_span,right_span=arg_right_span,case_sensitive=arg_case_sensitive,regexp=arg_regexp,show_progress=True)
                collocates.save_excel(self.workspace + self.corpus_in_use + '/output/collocates.xlsx')
                print('')
        except Exception as e:
            print(e)
            
    def do_collgraph(self,arg):
        parser = self.get_parser('collgraph')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_coll_pos = None 
                arg_case_sensitive = False
                arg_regexp = False
                arg_left_span = 5
                arg_right_span = 5
                if args.pos != None:
                    arg_pos = args.pos 
                if args.coll_pos !=None:
                    arg_coll_pos = args.coll_pos
                if args.case != None:
                    if args.case == 'y':
                        arg_case_sensitive = True
                    else: 
                        arg_case_sensitive = False
                if args.regexp != None:
                    if args.regexp == 'y':
                        arg_regexp = True
                    else:
                        arg_regexp = False
                if args.left_span != None:
                    arg_left_span = args.left_span
                if args.right_span != None:
                    arg_right_span = args.right_span 
                collocates = corpus.collocates(args.node,pos=arg_pos,coll_pos=arg_coll_pos,left_span=arg_left_span,right_span=arg_right_span,case_sensitive=arg_case_sensitive,regexp=arg_regexp,show_progress=True)
                collgraph = CollGraph(node=args.node)
                collgraph.plot_graphcoll(collocates)
                print('')
        except Exception as e:
            print(e)
            
    def do_wtfreq(self,arg):
        parser = self.get_parser('wtfreq')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_lowercase = True
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        arg_lowercase = False
                wtfreq = corpus.wtfreq(lowercase=arg_lowercase,show_progress=True)
                wtfreq.save_excel(self.workspace + self.corpus_in_use + '/output/wtfreq.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_wfreqinfiles(self,arg):
        parser = self.get_parser('wfreqinfiles')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_lowercase = True
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        arg_lowercase = False
                wfreqinfiles = corpus.wfreqinfiles(lowercase=arg_lowercase,show_progress=True)
                wfreqinfiles.save_excel(self.workspace + self.corpus_in_use + '/output/wfreqinfiles.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_clusters(self,arg):
        parser = self.get_parser('clusters')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_size = 3 
                arg_minfreq = 1 
                arg_minrange = 1
                arg_lowercase = True 
                arg_lowercase = True
                
                if args.pos !=None:
                    arg_pos = args.pos 
                if args.size !=None:
                    arg_size = args.size 
                if args.minfreq !=None:
                    arg_minfreq = args.minfreq 
                if args.minrange != None:
                    arg_minrange = args.minrange
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        args.lowercase = False 
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        arg_lowercase = False
                clusters = corpus.clusters(args.word,pos=arg_pos,size=arg_size,minfreq=arg_minfreq,minrange=arg_minrange,lowercase=arg_lowercase,show_progress=True)
                clusters.save_excel(self.workspace + self.corpus_in_use + '/output/clusters.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_ngrams(self,arg):
        parser = self.get_parser('ngrams')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_size = 3 
                arg_minfreq = 1 
                arg_minrange = 1
                arg_lowercase = True 
                arg_lowercase = True
                
                if args.pos !=None:
                    arg_pos = args.pos 
                if args.size !=None:
                    arg_size = args.size 
                if args.minfreq !=None:
                    arg_minfreq = args.minfreq 
                if args.minrange != None:
                    arg_minrange = args.minrange
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        args.lowercase = False 
                if args.lowercase != None:
                    if args.lowercase != 'y':
                        arg_lowercase = False
                ngrams = corpus.ngrams(pos=arg_pos,size=arg_size,minfreq=arg_minfreq,minrange=arg_minrange,lowercase=arg_lowercase,show_progress=True)
                ngrams.save_excel(self.workspace + self.corpus_in_use + '/output/ngrams.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_dispersion(self,arg):
        parser = self.get_parser('dispersion')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_pos = None
                arg_case_sensitive = False
                arg_regexp = False
                arg_limit = None
                if args.pos != None:
                    arg_pos = args.pos 
                if args.case != None:
                    if args.case == 'y':
                        arg_case_sensitive = True
                    else: 
                        arg_case_sensitive = False
                if args.regexp != None:
                    if args.regexp == 'y':
                        arg_regexp = True
                    else:
                        arg_regexp = False
                if args.limit != None:
                    arg_limit = int(args.limit)
                dispersion = corpus.dispersion(args.node,pos=arg_pos,case_sensitive=arg_case_sensitive,regexp=arg_regexp,limit=arg_limit,show_progress=True)
                dispersion.save_excel(self.workspace + self.corpus_in_use + '/output/dispersion.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_keywords_dispersion(self,arg):
        parser = self.get_parser('keywords_dispersion')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if self.corpus_in_use is not None:
                corpus_info = self.__corpus_info(self.corpus_in_use)
                corpus = Corpus(self.workspace,self.corpus_in_use,corpus_info['language'])
                # args
                arg_lowercase = True 
                arg_limit = 25
                if args.lowercase != None:
                    arg_lowercase = args.lowercase
                if args.limit != None:
                    arg_limit = int(args.limit)
                wordlist = corpus.wordlist(show_progress=True)
                keywords = corpus.keywords(wordlist,show_progress=True)
                del wordlist 
                keywords_dispersion = corpus.keywords_dispersion(keywords,lowercase=arg_lowercase,limit=arg_limit,show_progress=True)
                keywords_dispersion.save_excel(self.workspace + self.corpus_in_use + '/output/keywords_dispersion.xlsx')
                print('')
        except Exception as e:
            print(e)
    
    def do_text2utf8(self,arg):
        parser = self.get_parser('text2utf8')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            arg_source = args.source
            arg_target = args.target
            if os.path.exists(arg_source) == True:
                if os.path.isdir(arg_source) == True:
                    if os.path.exists(arg_target)==False:
                        os.mkdir(arg_target)
                    kit_util.files2utf8(arg_source, arg_target, 'mbcs', True)
                else:
                    print('The path is not a valid source folder or directory.')
            else:
                print('The path provided is not valid.')
        except Exception as e:
            print(e)
    
    def do_examples(self,arg):
        parser = self.get_parser('examples')
        args = parser.parse_args(self.parse_arg(arg))
        try:
            if args.dest_path !=None:
                ex = Examples()
                ex.download(dest_path=args.dest_path)
            else:
                ex = Examples()
                ex.download()
        except Exception as e:
            print(e)
        
    
    def run(self):
        print('Welcome to the Kitconc shell. Version: %s' % version.__version__)
        print('Type help or ? to list commands.\n')
        self.load_workspace()
        self.prompt = 'kitconc>'
        self.cmdloop('')
        
    

if __name__ == '__main__':
    Kit().run()






# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import sys 
from kitconc.kit_corpus import Corpus 
print('Hello!')
print('')
print("""This is an example on how to create a custom scripts.
\nHere is how arguments are received: 
""")
print('Arg 1 - Workspace: %s' % sys.argv[1])
print('Arg 2 - Corpus: %s' % sys.argv[2])
if len(sys.argv)>3:
    for i in range(3,len(sys.argv)):
        print('Arg %s: %s' % (i,sys.argv[i]))
print('\nHere is how to use the Corpus object:')
corpus = Corpus(sys.argv[1],sys.argv[2])
print('\nTokens:  %s' % corpus.tokens())


import os,sys 
import numpy as np 
import multiprocessing as mp 

def get_files ():
    path = sys.argv[1] + sys.argv[2] + '/data/temp/'
    for filename in os.listdir(path):
        yield filename 

def create_npy_array(filename):
    source_path = sys.argv[1] + sys.argv[2] + '/data/temp/' + filename 
    dest_path = sys.argv[1] + sys.argv[2] + '/data/tagged/' + filename 
    arr=np.loadtxt(source_path, dtype=int, delimiter='\t',encoding='utf-8')
    np.save(dest_path,arr,allow_pickle=False)
    
def create_npy_arrays():
    filenames = get_files()
    p = mp.Pool(mp.cpu_count())
    p.map(create_npy_array,filenames)
    p.close()
    p.join()

if __name__ == '__main__':
    create_npy_arrays()
    



import os 
import numpy as np
cimport numpy as np
cimport cython
DTYPE = np.int
ctypedef np.int_t DTYPE_t
import pickle 
import re 

cdef dict load_dict(str dict_path):
    cdef:
        dict d  
    with open(dict_path,'rb') as fh:
        d = dict(pickle.load(fh))
    return d 

cdef tuple parse_search(str workspace,str corpus_name,str node,object pos, object case_sensitive,object regexp):
    """Parses the input search and returns a tuple with encoded word and pos node.
    """
    cdef:
        dict dict_words
        dict dict_tags
        object k 
        object v 
        list n 
        str w 
        list p 
        str t 
        object encoded_node
        object encoded_pos
        list patterns 
        object ptrn
        np.ndarray arr
        list a
         
    # load dictionaries for encoding words to numbers
    if case_sensitive == False:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/wordslc.pickle') # lowercase
    else:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
        dict_words = {v: k for k, v in dict_words.items()}
    dict_tags = load_dict(workspace+corpus_name+'/data/idx/tags.pickle')
    dict_tags = {v: k for k, v in dict_tags.items()}
    # parse node
    n = []
    if case_sensitive == False:
        node = node.lower()
    for w in node.strip().split(' '):
        if len(w.strip()) != 0:
            n.append(w.strip())
    # parse pos 
    if pos != None:
        p = []
        if type(pos) == list:
            for t in pos:
                if len(t.strip())!= 0:
                    p.append(t.strip())
        else:
            for t in pos.strip().split(' '):
                if len(t.strip())!=0:
                    p.append(t.strip())
    # encode node
    encoded_node = []
    if regexp == False:
        for w in n:
            if w in dict_words:
                arr = np.array(dict_words[w],dtype=int)
                encoded_node.append(arr)
            else: # if any word has no match, return None
                encoded_node = None
                break
    else: # use regexp:
        patterns = []
        for ptrn in n:
            patterns.append(re.compile(ptrn))
        for ptrn in patterns:
            a = []
            for k in dict_words:
                if ptrn.match(k) != None:
                    a+=dict_words[k]
            if len(a)!=0:
                arr = np.array(a,dtype=int)
                encoded_node.append(arr)
            else: # if any pattern has no match, return None
                encoded_node = None
                break 
    # encode pos
    if pos != None:
        encoded_pos = []
        for t in p:
            if t in dict_tags:
                arr = np.array(dict_tags[t],dtype=int)
                encoded_pos.append(arr)
            else: # if any tag has no match, return None
                encoded_pos = None
                break 
    else:
        encoded_pos = None
    return (encoded_node, encoded_pos)

cdef list search_node(str npy_path, object encoded_node, object encoded_pos):
    """Searches the node in every text file and gets contexts."""
    cdef:
        list files 
        list sentences 
        int size 
        str filename 
        np.ndarray arr
        np.ndarray ix
    files = os.listdir(npy_path)
    sentences = []
    size = len(encoded_node)
    for filename in files:
        # load arr 
        arr = np.load(npy_path + filename)
        # get matching indexes
        ix = arr_search_indexes(arr,encoded_node,encoded_pos)
        # get sentences from indexes
        sentences += get_sentences(arr,ix,size)
    return sentences 



cdef np.ndarray arr_search_indexes(np.ndarray arr,object search_w,object search_t):
    """Searches the node in arr and returns the matching indexes. (Max. 3 words)"""
    cdef:
        int size
        np.ndarray ix 
        np.ndarray iw
        np.ndarray it 
        np.ndarray x1 
        np.ndarray x2 
        np.ndarray x3 
        np.ndarray t1
        np.ndarray t2
        np.ndarray t3
        np.ndarray x 
        np.ndarray y 
        np.ndarray z 
    size = len(search_w)
    if search_t == None:
        # search only words
        if size == 1:
            ix = np.isin(arr[:,[0]],search_w)
            ix = np.where(ix)[0]
        elif size == 2:
            # w1
            x1 = np.isin(arr[:,[0]],search_w[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            # w2 
            x2 = np.isin(arr[:,[0]],search_w[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-1)
            y = np.arange(1,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & x2[y])[0]
            return ix 
        elif size == 3:
            # w1
            x1 = np.isin(arr[:,[0]],search_w[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            # w2
            x2 = np.isin(arr[:,[0]],search_w[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            # w3
            x3 = np.isin(arr[:,[0]],search_w[2])
            if np.any(x3)==False:
                return np.where(x3)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-2)
            y = np.arange(1,x1.shape[0]-1)
            z = np.arange(2,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & x2[y] & x3[z])[0]
    else:
        # search words and POS
        if size == 1:
            iw = np.isin(arr[:,[0]],search_w)
            it = np.isin(arr[:,[1]],search_t)
            ix = np.where(iw & it)[0]
        elif size == 2:
            # w1 t1
            x1 = np.isin(arr[:,[0]],search_w[0])
            t1 = np.isin(arr[:,[1]],search_t[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            if np.any(t1)==False:
                return np.where(t1)[0]
            # w2 t2 
            x2 = np.isin(arr[:,[0]],search_w[1])
            t2 = np.isin(arr[:,[1]],search_t[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            if np.any(t2)==False:
                return np.where(t2)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-1)
            y = np.arange(1,x1.shape[0])
            # return indexes 
            ix = np.where(x1[x] & t1[x] & x2[y] & t2[y])[0]
        elif size == 3:
            # w1 t1
            x1 = np.isin(arr[:,[0]],search_w[0])
            t1 = np.isin(arr[:,[1]],search_t[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            if np.any(t1)==False:
                return np.where(t1)[0]
            # w2 t2
            x2 = np.isin(arr[:,[0]],search_w[1])
            t2 = np.isin(arr[:,[1]],search_t[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            if np.any(t2)==False:
                return np.where(t2)[0]
            # w3 t3
            x3 = np.isin(arr[:,[0]],search_w[2])
            t3 = np.isin(arr[:,[1]],search_t[2])
            if np.any(x3)==False:
                return np.where(x3)[0]
            if np.any(t3)==False:
                return np.where(t3)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-2)
            y = np.arange(1,x1.shape[0]-1)
            z = np.arange(2,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & t1[x] & x2[y] & t2[y] & x3[z] & t3[z])[0]
    return ix 

cdef tuple get_sentence(np.ndarray arr, int sent_id, object i):
    cdef:
        tuple ix 
        int token_id
        np.ndarray arr_sent 
    ix = np.where(np.isin(arr[:,2],sent_id))
    token_id = np.where(np.isin(ix,i))[1][0]
    arr_sent = arr[ix][:,0]
    return (token_id,arr_sent.tolist())
    
cdef list get_sentences(np.ndarray arr,np.ndarray indexes, int size):
    cdef:
        list sentences
        object i 
        int sent_id 
        int token_id
        list sent  
        int t 
        int s 
        int f 
    sentences = list()
    for i in indexes:
        sent_id = arr[i,2]
        token_id,sent = get_sentence(arr,sent_id,i) # sent
        t = token_id + size      # token id
        s = arr[i,2]             # sentence id
        f = arr[i,3]             # file id
        sentences.append((sent,t,s,f))
    return sentences 

cdef list translate(list sentences,dict dict_words,dict dict_files, int size):
    """Translate numbers back to words in a list format for kitconc KIWC object."""
    cdef:
        list concordance 
        int i  
        list c 
        tuple sent 
    dict_files = {v: k for k, v in dict_files.items()} # reverse keys
    concordance = []
    i = 0
    for sent in sentences:
        i+=1
        c = []
        # concordance
        for i in sent[0]:
            c.append(dict_words[i])
        # add to list
        concordance.append((i, ' '.join(c), dict_files[sent[3]][0],sent[1]-(size-1),str(sent[2]),str(sent[3])))
    return concordance

def make_concordance(str workspace,str corpus_name, str node,object pos, object case_sensitive,object regexp):
    cdef:
        object encoded_node
        object encoded_pos 
        int nl 
        list sentences 
        dict dict_words 
        dict dict_files 
        list concordance 
    encoded_node,encoded_pos = parse_search(workspace,corpus_name,node,pos,case_sensitive,regexp)
    if encoded_node != None: # search node is OK
        nl = len(encoded_node)
        sentences = search_node(workspace+corpus_name + '/data/npy/',encoded_node,encoded_pos)
        if len(sentences)!=0:
            dict_words = load_dict(workspace+corpus_name + '/data/idx/words.pickle')
            dict_files = load_dict(workspace+corpus_name + '/data/idx/filenames.pickle')
            concordance = translate(sentences,dict_words,dict_files,nl)
        else:
            concordance=[]
    else: # no match in search node, set empty
        nl = 0
        concordance = []
    return (nl,concordance) 
        



import os,sys 
import numpy as np 
import pickle 
import multiprocessing as mp 

class Indexing(object):
    
    def __init__(self):
        self.workspace = sys.argv[1]
        self.corpus_name = sys.argv[2]
        self.language = sys.argv[3]
        self.dic_w = self.load_dict(self.workspace+self.corpus_name + '/data/idx/words.pickle')
        self.dic_t = self.load_dict(self.workspace+self.corpus_name + '/data/idx/tags.pickle')

    
    
    def get_files (self,source_folder):
        for filename in os.listdir(source_folder):
            yield filename

    def load_dict(self,path):
        with open(path,'rb') as fh:
            return pickle.load(fh)
    
    def save_npy(self,filename):
        d = []
        with open (self.workspace + self.corpus_name + '/data/tmp1/' + filename,'r', encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip())!= 0:
                    f = line.strip().split('\t')
                    if len(f) == 4:
                        d.append((self.dic_w[f[0]],self.dic_t.get(f[1],0),f[2],f[3]))
                    else:
                        print('Warning: cannot parse token in file %s ' % filename )
        np.save(self.workspace + self.corpus_name + '/data/npy/' + filename, np.array(d,dtype=int), allow_pickle=False)
    
    def save_inverted_dicts(self):
        self.dic_w = {v: k for k, v in self.dic_w.items()}
        with open(self.workspace + self.corpus_name + '/data/idx/words.pickle','wb') as fh:
            pickle.dump(self.dic_w,fh)
        self.dic_t = {v: k for k, v in self.dic_t.items()}
        with open(self.workspace + self.corpus_name + '/data/idx/tags.pickle','wb') as fh:
            pickle.dump(self.dic_t,fh)
    
    def save_lc_dict(self):
        d = dict()
        for k,v in self.dic_w.items():
            key = k.lower()
            if key in d:
                d[key]+=[v]
            else:
                d[key]=[v]
        with open(self.workspace + self.corpus_name + '/data/idx/wordslc.pickle','wb') as fh:
            pickle.dump(d,fh)
            
        

    def execute(self):
        files = self.get_files(self.workspace + self.corpus_name + '/data/tmp1/')
        p = mp.Pool(mp.cpu_count())
        p.map(self.save_npy,files)
        p.close()
        p.join()
        self.save_lc_dict()
        self.save_inverted_dicts()


if __name__ == '__main__':
    idx = Indexing()
    idx.execute()


                
        
    
    
     



# Author: Hayk Aleksanyan
# color related issued of word clouds are handled here

import random


def hslToRgb(h, s, l):
    """
       convert from HSL (hue-saturation-luminosity) to RGB format
       h, s, l are all in the range [0,1]
       r, g, b will be in the range [0,255]
       see https://stackoverflow.com/questions/2353211/hsl-to-rgb-color-conversion
    """

    r, g, b =  0, 0, 0

    if s == 0:
        r, g, b = l, l, l
    else:
        def converter( p, q, t ):
            if t < 0:
                 t += 1
            if t > 1:
                t -= 1
            if t < 1/6:
                 return p + (q - p) * 6 * t
            if t < 1/2:
                return q
            if t < 2/3:
                return p + (q - p) * (2/3 - t) * 6
            return p

        q = l * (1 + s) if l < 0.5 else l + s - l * s
        p = 2 * l - q
        r = converter(p, q, h + 1/3)
        g = converter(p, q, h)
        b = converter(p, q, h - 1/3)

    return int(r * 255), int(g * 255), int(b * 255)

def getRandomColor(h = -1, s = -1, l = -1, bright = False):
    """
        return a random RGB triple which is not very bright as a color, if @bright == False; and is bright otherwise
        we allow for some of the components of the HSL be predefined, and others being chosen at random
        to fix the specific HSL coordinate change the -1 value to the one you wish from interval [0,1]

        Note ==>  bright colors need high saturation and high luminosity
        Note ==>  lower saturation means grayish colors
    """

    if h == -1 or not 0 <= h <= 1:
        h = random.random()
    if s == -1 or not 0 <= s <= 1:
        s = random.random()
        if bright == False:
            if s > 0.8:
                s = 0.5
        else:
            if s < 0.5:
                s = 1 - s
    if l == -1 or not 0 <= l <= 1:
        l = random.random()
        if bright == False:
            if l > 0.5:
                l = 1 - l
        else:
            if l < 0.5:
                l = 1 - l

    return hslToRgb(h, s, l)



def colorTokens(normalTokens, background = 0):
    """
      given a list of Tokens (see the wordle for Token class) we add their color attributes
      @background == 0 means WHITE; == 1 means Black background, needs brighter colors
      by picking a random color scheme and applying on the Tokens
    """

    if background == 0:
        scheme = random.randint(0, 3)
        if scheme == 0:
            print('\nColor scheme: Random\n')
            randomColors(normalTokens)
        elif scheme == 1:
            print('\nColor scheme: Jet\n')
            jetColors(normalTokens)
        elif scheme == 2:
            print('\nColor scheme: Grayish\n')
            grayishRandomColors(normalTokens)
        elif scheme == 3:
            print('\nColor scheme: From a fixed list\n')
            chooseFromFixedSchemes(normalTokens)
    else:
            print('\nColor scheme: Random, black background\n')
            randomColors(normalTokens, background = 1)



def randomColors(normalTokens, background = 0):
    # apply random colors of slightly darker gamma on white background == 0
    # and bright colors on black background == 1
    for token in normalTokens:
        if background == 0:
            token.color = getRandomColor()
        else:
            token.color = getRandomColor(h = -1, s = -1, l = -1, bright = True)


def jetColors(normalTokens):
    # apply jet color (heatmap) scheme on tokens
    if not normalTokens:  return

    h, s, l = 0, 0.85, 0.4
    step = 0.7/len(normalTokens)

    for token in normalTokens:
        token.color = hslToRgb(h, s, l)
        h += step

def grayishRandomColors(normalTokens):
    # apply grayscale random colors
    for token in normalTokens:
        # get a random color with lower saturation, this will ensure a grayish color
        token.color = getRandomColor(h = -1, s = 0, l = -1)

def chooseFromFixedSchemes(normalTokens):
    # choose a color scheme from a fixed list which colors tokens based on the fontSize

    color_schemes = [
        [ (89, 97, 113), (115, 124, 140), (141, 150, 168), (179, 188, 204), (219, 227, 240) ],
        [ (89, 97, 113), (115, 124, 140), (141, 150, 168) , (89, 97, 113), (89, 97, 113)],
        [ (120, 120, 120), (0,100,149),  (242, 99, 95), (0, 76, 112), (244,208,12) ],
        [ (14, 38, 50), (1,70,99),  (35, 118, 150), (180, 200, 207), (159,195,185) ],
        [ (3, 113, 146), (99,167,190),  (10, 31, 78), (252, 105, 53), (252,105,53) ],
        [ (12, 6, 54), (9,81,105),  (5, 155, 154), (83, 186, 131), (159,217,107) ] ,
        [ (100, 101, 165), (105,117,167),  (244, 233, 109), (242, 138, 49), (241,88,56) ],
        [ (171, 165, 191), (90,87,118),  (88, 62, 47), (241, 224, 214), (191,153,144) ],
        [ (25, 46, 91), (30,101,167),  (115, 162, 192), (0, 116, 63), (241,161,4) ]
    ]

    # chose the color scheme randomly
    word_colors = color_schemes[ random.randint(0, len(color_schemes) - 1) ]
    max_size = max( [ token.fontSize for token in normalTokens ] )

    for token in normalTokens:
        if token.fontSize >= 0.7*max_size:
            c = word_colors[-1]
        elif ((token.fontSize >= 0.5*max_size)and(token.fontSize < 0.7*max_size) ):
            c = word_colors[-2]
        elif ((token.fontSize >= 0.35*max_size)and(token.fontSize < 0.5*max_size) ):
            c = word_colors[-3]
        elif ((token.fontSize >= 0.15*max_size)and(token.fontSize < 0.35*max_size) ):
            c = word_colors[-4]
        else:
            c = word_colors[-5]

        token.color = c

def chooseFromCustom(normalTokens):
    # choose a color scheme from a fixed list which colors tokens based on the fontSize

    color_schemes = [
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197) ], # BRFLAG
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197)],
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197)],
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197)],
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197)],
        [ (45, 83, 148), (78, 142, 41), (231, 243, 65), (137, 239, 195), (239, 247, 197) ]
    ]

    # chose the color according to index
    word_colors = color_schemes[0]
    max_size = max( [ token.fontSize for token in normalTokens ] )

    for token in normalTokens:
        if token.fontSize >= 0.7*max_size:
            c = word_colors[-1]
        elif ((token.fontSize >= 0.5*max_size)and(token.fontSize < 0.7*max_size) ):
            c = word_colors[-2]
        elif ((token.fontSize >= 0.35*max_size)and(token.fontSize < 0.5*max_size) ):
            c = word_colors[-3]
        elif ((token.fontSize >= 0.15*max_size)and(token.fontSize < 0.35*max_size) ):
            c = word_colors[-4]
        else:
            c = word_colors[-5]

        token.color = c



import os
import numpy as np
cimport numpy as np
cimport cython
DTYPE = np.int
ctypedef np.int_t DTYPE_t 
import pickle 

def get_files (str source_folder):
    cdef object filename 
    for filename in os.listdir(source_folder):
        yield filename


cdef dict load_dict(str path):
    with open(path,'rb') as fh:
        return pickle.load(fh)

def make_indexes(str workspace, str corpus_name, str language):
    cdef str source_path 
    cdef object files
    cdef dict dic_w
    cdef dict dic_t 
    cdef list d 
    cdef str line 
    cdef str k 
    cdef int v  
    source_path = workspace + corpus_name + '/data/tmp1/'
    files = get_files(source_path)
    dic_w = load_dict(workspace+corpus_name + '/data/idx/words.pickle')
    dic_t = load_dict(workspace+corpus_name + '/data/idx/tags.pickle')
    for filename in files:
        d = list()
        with open (source_path + filename,'r', encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip())!= 0:
                    f = line.strip().split('\t')
                    d.append((dic_w[f[0]],dic_t[f[1]],f[2],f[3]))
        np.save(workspace + corpus_name + '/data/npy/' + filename, np.array(d,dtype=int), allow_pickle=False)
    
    dic_w = {v: k for k, v in dic_w.items()}
    with open(workspace + corpus_name + '/data/idx/words.pickle','wb') as fh:
        pickle.dump(dic_w,fh)
    
    dic_t = {v: k for k, v in dic_t.items()}
    with open(workspace + corpus_name + '/data/idx/tags.pickle','wb') as fh:
        pickle.dump(dic_t,fh)
                
        
    
    
     



# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os,sys   
import time  
from kitconc import kit_tools
from kitconc import kit_util
from kitconc import kit_data 
import subprocess
import pandas as pd
import collections 
# cython 
try:
    from cython import __version__
    USING_CYTHON = True
except:
    USING_CYTHON = False

#USING_CYTHON = False

if USING_CYTHON == True:
    from kitconc.cy_wordlist import make_wordlist
    from kitconc.cy_keywords import make_keywords
    from kitconc.cy_wtfreq import make_wtfreq
    from kitconc.cy_wfreqinfiles import make_wfreqinfiles
    from kitconc.cy_kwic import make_kwic 
    from kitconc.cy_concordance import make_concordance
    from kitconc.cy_collocates import make_collocates
    from kitconc.cy_clusters import make_clusters     
    from kitconc.cy_ngrams import make_ngrams
    from kitconc.cy_dispersion import make_dispersion
    from kitconc.cy_keywords_dispersion import make_keywords_dispersion
else:
    from kitconc.py_wordlist import make_wordlist
    from kitconc.py_keywords import make_keywords
    from kitconc.py_wtfreq import make_wtfreq
    from kitconc.py_wfreqinfiles import make_wfreqinfiles
    from kitconc.py_kwic import make_kwic 
    from kitconc.py_concordance import make_concordance
    from kitconc.py_collocates import make_collocates
    from kitconc.py_clusters import make_clusters     
    from kitconc.py_ngrams import make_ngrams
    from kitconc.py_dispersion import make_dispersion
    from kitconc.py_keywords_dispersion import make_keywords_dispersion
    print('Warning: Could not use cython or error in pyx file.')
    

    

class Corpus (object):
    """This class is the main component for doing corpus analysis in Kitconc."""
        
    def __init__(self,workspace,corpus_name,language='english',**kwargs):
        """
        :param  workspace: A folder path where the corpus will be stored
        :type   workspace: str
        :param corpus_name: A name for corpus identification. 
        :type corpus_name: str
        :param language: The language for text processing. 
        :type language: str 
        """
        # script path
        self.__path = os.path.dirname(os.path.abspath(__file__))
        self.resource_data_path = self.__path + '/data/'
        # check environment
        self.__shell = False
        try:
            from IPython import get_ipython
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                self.__shell = True   # Jupyter notebook or qtconsole
        except:
            pass  
        # constants
        self.MEASURE_MUTUAL_INFORMATION = 'mutual information'
        self.MEASURE_CHI_SQUARE = 'chi-square'
        self.MEASURE_LOGLIKELIHOOD = 'log-likelihood'
        self.MEASURE_TSCORE = 't-score'
        # normalize workspace path 
        if str(workspace).endswith('/'):
            self.workspace = workspace
        else:
            self.workspace = workspace + '/'
        # check workspace path exists
        # if not create it 
        if os.path.exists(self.workspace) == False:
            os.mkdir(self.workspace)
        # check corpus already exists
        if os.path.exists(self.workspace + corpus_name + '/info.tab'):
            # read info
            info = self.__load_corpus_info(self.workspace, corpus_name)
            self.corpus_name = corpus_name 
            self.language = info['language']
            self.encoding = info['encoding']
            self.output_path = self.workspace + self.corpus_name + '/output/'
            self.data_path = self.workspace + self.corpus_name + '/data/'
            self.tagged_path = self.workspace + self.corpus_name + '/data/tagged/'
        else:
            # get kwargs
            encoding = 'utf-8' # default
            # set variables 
            self.corpus_name = corpus_name
            self.language = language 
            self.encoding = encoding
            self.output_path = self.workspace + self.corpus_name + '/output/'
            self.data_path = self.workspace + self.corpus_name + '/data/'
            self.tagged_path = self.workspace + self.corpus_name + '/data/tagged/'
            # create corpus folder
            if not os.path.exists(self.workspace + self.corpus_name):
                os.mkdir(self.workspace + self.corpus_name)
            if not os.path.exists(self.workspace + self.corpus_name + '/data'):
                os.mkdir(self.workspace + self.corpus_name + '/data')
            if not os.path.exists(self.workspace + self.corpus_name + '/output'):
                os.mkdir(self.workspace + self.corpus_name + '/output')
            # make dirs for processing
            if not os.path.exists(self.workspace + self.corpus_name + '/data/npy'):
                os.mkdir(self.workspace + self.corpus_name + '/data/npy')
            if not os.path.exists(self.workspace + self.corpus_name + '/data/idx'):
                os.mkdir(self.workspace + self.corpus_name + '/data/idx')
            if not os.path.exists(self.workspace + self.corpus_name + '/data/tmp1'):
                os.mkdir(self.workspace + self.corpus_name + '/data/tmp1')
            if not os.path.exists(self.workspace + self.corpus_name + '/data/tmp2'):
                os.mkdir(self.workspace + self.corpus_name + '/data/tmp2')
            if not os.path.exists(self.workspace + self.corpus_name + '/data/tmp3'):
                os.mkdir(self.workspace + self.corpus_name + '/data/tmp3')
            
            
            # save info
            self.__save_corpus_info(self.workspace, self.corpus_name, self.language, self.encoding)
           
    # corpus info
    
    def __save_corpus_info(self,workspace,corpus_name,language,encoding):
        info = []
        info.append('workspace:' + '\t' + workspace)
        info.append('corpus name:' + '\t' + corpus_name)
        info.append('language:' + '\t' + language)
        info.append('encoding:' + '\t' + encoding)
        with open(workspace + corpus_name + '/info.tab','w',encoding='utf-8') as fh:
            fh.write('\n'.join(info))
    
    def __load_corpus_info (self,workspace,corpus_name):
        info = {}
        with open(workspace + corpus_name + '/info.tab','r',encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip()) !=0:
                    fields = line.strip().split('\t')
                    if len(fields) >= 2:
                        info[fields[0].replace(':','')]=fields[1]
        return info
    
    # show progress function
    
    def __progress(self, count, total, suffix=''):
        if self.__shell == False:
            bar_len = 30
            filled_len = int(round(bar_len * count / float(total)))
            percents = round(100.0 * count / float(total), 1)
            bar = '=' * filled_len + '-' * (bar_len - filled_len)
            sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', suffix))
            try:
                sys.stdout.flush()  
            except:
                pass 
  

    def add_texts(self,source_folder,**kwargs):
        """Adds texts to a corpus reference from a source folder.
        :param  source_folder: A folder path where the source texts are stored.
        :type   source_folder: str
        :param tagged: If texts are already tagged. 
        :type tagged: boolean
        :param language_model: A language model for tokenization and tagging texts.
        :type language_model: object
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        """
        import shutil
        # args 
        tagged = kwargs.get('tagged',False)
        language_model = kwargs.get('language_model',None)
        show_progress = kwargs.get('show_progress',False)
        if not source_folder.endswith('/'):
            source_folder = source_folder + '/'
        # time it start
        if show_progress == True:
            t0 = time.time()
            print('New corpus:')
        # make indexes
        if language_model == None:
            # check if there's a saved language model
            has_saved_model = False 
            lm_path = self.__path + '/data/' + self.language + '_model.pickle'
            if os.path.exists(lm_path):
                has_saved_model = True
            # for sepated and language models 
            if tagged == False:
                if has_saved_model == False:
                    # use separate saved models
                    if show_progress == True:
                        print('Tagging...')
                    resources_path = self.__path + '/data/'
                    subprocess.call(['python',self.__path +'/tagging.py', resources_path,self.workspace,self.corpus_name, self.language,source_folder])
                    if show_progress == True:
                        print('Making indexes...')
                    subprocess.call(['python',self.__path +'/indexing.py', self.workspace,self.corpus_name, self.language])
                else:
                    # use a saved model
                    # saved models must be in kitconc data folder
                    if show_progress == True:
                        print('Tagging...')
                    # create temp folder
                    if os.path.exists(self.workspace+self.corpus_name+'/data/tmp0'):
                        shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp0/')
                    os.mkdir(self.workspace+self.corpus_name+'/data/tmp0/')
                    # load model 
                    import pickle 
                    with open (self.__path + '/data/' + self.language + '_model.pickle','rb') as fh:
                        language_model = pickle.load(fh)
                    # tag texts
                    files = os.listdir(source_folder)
                    for filename in files:
                        with open(source_folder + filename,'r',encoding='utf-8') as fh:
                            text = fh.read()
                        doc = language_model(text)
                        tagged_sents = []
                        for sent in doc.sents:
                            tagged_sents.append( ' '.join( [(token.text + '/' + token.pos_) for token in sent if len(token.text.strip()) != 0]))
                        with open(self.workspace+self.corpus_name+'/data/tmp0/' + filename,'w',encoding='utf-8') as fh:
                            fh.write('\n'.join(tagged_sents)) 
                    # load tagged texts    
                    if show_progress == True:
                        print('Loading tagged corpus...')
                    resources_path = self.__path + '/data/'
                    subprocess.call(['python',self.__path +'/tagged.py', resources_path,self.workspace,self.corpus_name, self.language,self.workspace+self.corpus_name+'/data/tmp0/'])
                    if show_progress == True:
                        print('Making indexes...')
                    subprocess.call(['python',self.__path +'/indexing.py', self.workspace,self.corpus_name, self.language])
                    # remove temp folder
                    shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp0/')
                    
            else:
                if show_progress == True:
                    print('Loading tagged corpus...')
                resources_path = self.__path + '/data/'
                subprocess.call(['python',self.__path +'/tagged.py', resources_path,self.workspace,self.corpus_name, self.language,source_folder])
                if show_progress == True:
                    print('Making indexes...')
                subprocess.call(['python',self.__path +'/indexing.py', self.workspace,self.corpus_name, self.language])
        else:
            # use language model on the fly to tag texts
            # it allows to use spaCy
            if show_progress == True:
                print('Tagging...')
            # create temp folder
            if os.path.exists(self.workspace+self.corpus_name+'/data/tmp0'):
                shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp0/')
            os.mkdir(self.workspace+self.corpus_name+'/data/tmp0/')
            # tag texts
            files = os.listdir(source_folder)
            for filename in files:
                with open(source_folder + filename,'r',encoding='utf-8') as fh:
                    text = fh.read()
                doc = language_model(text)
                tagged_sents = []
                for sent in doc.sents:
                    tagged_sents.append( ' '.join( [(token.text + '/' + token.pos_) for token in sent if len(token.text.strip()) != 0]))
                with open(self.workspace+self.corpus_name+'/data/tmp0/' + filename,'w',encoding='utf-8') as fh:
                    fh.write('\n'.join(tagged_sents)) 
            # load tagged texts    
            if show_progress == True:
                print('Loading tagged corpus...')
            resources_path = self.__path + '/data/'
            subprocess.call(['python',self.__path +'/tagged.py', resources_path,self.workspace,self.corpus_name, self.language,self.workspace+self.corpus_name+'/data/tmp0/'])
            if show_progress == True:
                print('Making indexes...')
            subprocess.call(['python',self.__path +'/indexing.py', self.workspace,self.corpus_name, self.language])
            # remove temp folder
            shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp0/')
        
        # add source path to info
        ntexts = len(os.listdir(source_folder))
        with open(self.workspace+self.corpus_name+'/info.tab','a',encoding='utf-8') as fh:
            fh.write('\nTextfiles:\t%s\nSource:\t%s' % (ntexts,source_folder))
        # time it end 
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('Total time: %s seconds' % total_time)
        # delete temporary folders
        shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp1/')
        shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp2/')
        shutil.rmtree(self.workspace+self.corpus_name+'/data/tmp3/')
    

    def wordlist(self,**kwargs):
        """
        Generates a frequency wordlist based on text files.
        :param lowercase: Letters are converted to lowercase 
        :type lowercase: boolean
        :param min_freq: minimum frequency of the word
        :type min_freq: int
        :param show_progress: Prints a progress message if value is True 
        :type show_progress: boolean
        :return: Wordlist
        :rtype: Wordlist
        """
        # args
        show_progress=kwargs.get('show_progress',False)
        min_freq = kwargs.get('min_freq',1)
        lowercase = kwargs.get('lowercase',True)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make
        tokens, types, type_token, hapax, xwordlist = make_wordlist(self.workspace,self.corpus_name, self.language,lowercase)
        wlst = kit_tools.Wordlist(tokens=tokens,types=types,typetoken=round(type_token,2),hapax=hapax)
        wlst.df = pd.DataFrame(xwordlist,columns=['N','WORD','FREQUENCY','%'])
        wlst.df=wlst.df[(wlst.df.FREQUENCY >= min_freq)] 
        wlst.df['%'] = wlst.df['%'].apply(lambda x: round(x, 2))
        # time end
        if show_progress == True:
            #self.__progress(types, types, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return  wlst
    
    def keywords(self,**kwargs):
        """
        Extracts keywords from a wordlist
        :param measure: A statistical measure to use (log-likelihood or chi square)
        :type measure : str
        :param stoplist: a list of words to be filtered out
        :type stoplist: list
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Keywords
        :rtype: Keywords
        """
        #args
        measure = kwargs.get('measure','log-likelihood')
        stoplist = kwargs.get('stoplist',[])
        show_progress=kwargs.get('show_progress',False)
        # get statistics measures
        if measure == 'log-likelihood':
            stat = 1
        elif measure == 'chi-square':
            stat = 2
        else:
            stat = 1
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make keywords table
        kwlst = kit_tools.Keywords()
        k = make_keywords(self.workspace,self.corpus_name,self.language,stat)
        kwlst.df = pd.DataFrame(k,columns=['N','WORD','FREQUENCY','KEYNESS'])
        k = None
        # use stoplist
        kwlst.df = kwlst.df[~kwlst.df['WORD'].isin(stoplist)]
        # sort by keyness and reset index
        kwlst.df['KEYNESS'] = kwlst.df['KEYNESS'].apply(lambda x: round(x, 2))
        kwlst.df.sort_values('KEYNESS',ascending=False,inplace=True)
        kwlst.df.N = [(i+1)for i in range(0,len(kwlst.df))]
        kwlst.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time) 
        return kwlst
    
    def wtfreq(self,**kwargs):
        """
        Generates a word tag frequency list based on corpus text files.
        :param lowercase: Letters are converted to lowercase
        :type lowercase : boolean
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: WTfreq
        :rtype: WTfreq
        """
        #args
        lowercase = kwargs.get('lowercase',True)
        show_progress=kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make wtfreq
        lst_wt = make_wtfreq(self.workspace, self.corpus_name, self.language, lowercase)
        #s.append("N\tWORD\tTAG\tFREQUENCY\t%")
        wt = kit_tools.WTfreq()
        wt.df = pd.DataFrame(lst_wt,columns=['N','WORD','TAG','FREQUENCY','%'])
        # sort by freq and reset index
        wt.df['%'] = wt.df['%'].apply(lambda x: round(x, 2))
        wt.df.sort_values('FREQUENCY',ascending=False,inplace=True)
        wt.df.N = [(i+1)for i in range(0,len(wt.df))]
        wt.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            self.__progress(100,100, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return wt
    
    def wfreqinfiles(self,**kwargs):
        """
        Generates a frequency list based on word occurrence in corpus text files.
        :param lowercase: Letters are converted to lowercase
        :type lowercase : boolean
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Wfreqinfiles
        :rtype: Wfreqinfiles
        """
        #args
        lowercase = kwargs.get('lowercase',True)
        show_progress=kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # process
        lst_fif = make_wfreqinfiles(self.workspace, self.corpus_name, self.language, lowercase)
        fif = kit_tools.Wfreqinfiles()
        fif.df = pd.DataFrame(lst_fif,columns=['N','WORD','RANGE','%'])
        lst_fif = None
        # sort by freq and reset index
        fif.df['%'] = fif.df['%'].apply(lambda x: round(x, 2))
        fif.df.sort_values('RANGE',ascending=False,inplace=True)
        fif.df.N = [(i+1)for i in range(0,len(fif.df))]
        fif.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            self.__progress(100,100, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return fif
    
    def kwic(self,node,**kwargs):
        """
        Generates concordance lines.
        :param node: Search word or phrase (max. 4 words)
        :type node : str
        :param pos: Part of speech (POS) for each search word 
        :type pos: str or list
        :param case_sensitive: Enable regular expression for search word matching
        :type case_sensitive: boolean
        :param regexp: Enable regular expression for search matching
        :type regexp: boolean
        :param horizon: Left and right horizon of words
        :type horizon: int
        :param limit: Number of concordance lines to return
        :type limit: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Kwic
        :rtype: Kwic
        """
        # args
        node = node.strip()
        pos = kwargs.get('pos',None)
        case_sensitive = kwargs.get('case_sensitive',False)
        regexp = kwargs.get('regexp',False)
        horizon = kwargs.get('horizon',10)
        limit = kwargs.get('limit',None)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # deal with args
        nl,lst_kwic = make_kwic(self.workspace, self.corpus_name,node,pos,case_sensitive,regexp,horizon)
        kwic = kit_tools.Kwic(node_length=nl)
        kwic.df = pd.DataFrame(lst_kwic,columns=['N','LEFT','NODE','RIGHT','FILENAME','TOKEN_ID','SENT_ID','FILE_ID'])
        lst_kwic = None
        # limit
        if limit != None:
            kwic.df = kwic.df.head(limit)
        # reset indexes 
        kwic.df.N = [(i+1)for i in range(0,len(kwic.df))]
        kwic.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            self.__progress(100,100, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return kwic 
    
    def concordance(self,node,**kwargs):
        """
        Generates concordance lines.
        :param node: Search word or phrase (max. 4 words)
        :type node : str
        :param pos: Part of speech (POS) for each search word 
        :type pos: str or list
        :param case_sensitive: Enable regular expression for search word matching
        :type case_sensitive: boolean
        :param regexp: Enable regular expression for search matching
        :type regexp: boolean
        :param limit: Number of concordance lines to return
        :type limit: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Concordance
        :rtype: Concordance
        """
        # args
        node = node.strip()
        pos = kwargs.get('pos',None)
        case_sensitive = kwargs.get('case_sensitive',False)
        regexp = kwargs.get('regexp',False)
        limit = kwargs.get('limit',None)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # deal with args
        nl,lst_conc = make_concordance(self.workspace, self.corpus_name,node,pos,case_sensitive,regexp)
        conc = kit_tools.Concordance(node_length=nl)
        conc.df = pd.DataFrame(lst_conc,columns=['N','CONCORDANCE','FILENAME','TOKEN_ID','SENT_ID','FILE_ID'])
        lst_conc = None
        # limit
        if limit != None:
            conc.df = conc.df.head(limit)
        # reset indexes 
        conc.df.N = [(i+1)for i in range(0,len(conc.df))]
        conc.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            self.__progress(100,100, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return conc 
    
    def collocates(self,node,**kwargs):
        """
        Extracts a list of collocates.
        :param node: search word
        :type node : str
        :param pos: Part of speech (POS) for each search word 
        :type pos: str or list
        :param coll_pos: POS for filtering collocates
        :type coll_pos: str or list
        :param case_sensitive: Enable regular expression for search word matching
        :type case_sensitive: boolean
        :param regexp: Enable regular expression for search matching
        :type regexp: boolean
        :param left_span: Left and right horizon of words
        :type left_span: int
        :param right_span: Left and right horizon of words
        :type right_span: int
        :param lowercase: Letters are converted to lowercase 
        :type lowercase: boolean
        :param measure: Statistic measure tscore or mutual information (default: tscore)
        :type measure: str
        :param limit: Number of result lines to return
        :type limit: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Collocates
        :rtype: Collocates
        """
        # args
        node = node.strip()
        pos = kwargs.get('pos',None)
        coll_pos = kwargs.get('coll_pos',None)
        case_sensitive = kwargs.get('case_sensitive',False)
        regexp = kwargs.get('regexp',False)
        left_span = kwargs.get('left_span',5)
        right_span = kwargs.get('right_span',5)
        lowercase = kwargs.get('lowercase',True)
        measure = kwargs.get('measure',1)
        limit = kwargs.get('limit',None)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make
        if measure == 't-score':
            measure = 1
        elif measure == 'mutual information':
            measure = 2
        lst_coll = make_collocates(self.workspace, self.corpus_name, node, pos, case_sensitive, regexp, coll_pos, lowercase, left_span, right_span,measure)
        coll = kit_tools.Collocates()
        coll.df = pd.DataFrame(lst_coll,columns=['N','WORD','FREQUENCY','LEFT','RIGHT','ASSOCIATION'])
        lst_coll = None
        # sort 
        coll.df.sort_values('ASSOCIATION',ascending=False,inplace=True)
        # limit
        if limit != None:
            coll.df = coll.df.head(limit)
        # reset indexes 
        coll.df['ASSOCIATION'] = coll.df['ASSOCIATION'].apply(lambda x: round(x, 2))
        coll.df.N = [(i+1)for i in range(0,len(coll.df))]
        coll.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            self.__progress(100,100, '')
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return coll
        
    
    def compared_collocates(self,coll1,coll2,**kwargs):
        """
        Compares two sets of collocates.
        :param coll1: Collocates set 1
        :type coll1: Collocates 
        :param coll2: Collocates set 2
        :type coll2: Collocates
        :param stat_cutoff: Statistic measure cuttoff (default: 0)
        :type stat_cutoff: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: ComparedCollocates
        :rtype: ComparedCollocates
        """
        # kwargs
        stat_cutoff = kwargs.get('stat_cutoff',0)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # get all data in dictionaries
        # and filter association measures with stat_cutoff:
        # words = all words from col1 and col2
        # w1 = col1 data dict
        # w2 = col2 data dict
        words = collections.defaultdict()
        w1 = collections.defaultdict()
        for row in coll1.df.itertuples(index=False):
            if row[5] >= stat_cutoff:
                k = row[1]
                v = (row[2],row[5])
                w1[k]=v
                if k not in words:
                    words[k]= 0
        del coll1 
        w2 = collections.defaultdict()
        for row in coll2.df.itertuples(index=False):
            if row[5] >= stat_cutoff:
                k = row[1]
                v = (row[2],row[5])
                w2[k]=v
                if k not in words:
                    words[k]= 0
        del coll2
        # put all data together,
        # calculate percent difference
        # and make table   
        s = []
        s.append('N\tWORD\tFREQ1\tFREQ2\tASSOCIATION1\tASSOCIATION2\tDIFFERENCE')
        i=0
        total = len(words)
        for w in sorted(words):
            i+=1
            # get data
            if w in w1:
                w1_f = w1[w][0]
                w1_am = w1[w][1]
            else:
                w1_f = 0
                w1_am = 0.0
            if w in w2:
                w2_f = w2[w][0]
                w2_am = w2[w][1]
            else:
                w2_f = 0
                w2_am = 0.0
            # calcucale percent difference
            if w1_am <= w2_am: # ensure n2 is larger
                n1 = w1_am 
                n2 = w2_am 
            else:
                n1 = w2_am 
                n2 = w1_am 
            pd = round(((n2-n1) / float(n2)) * 100,2)
            # add to table
            s.append('\t'.join([str(i), str(w),str(w1_f),str(w2_f),str(w1_am),str(w2_am),str(pd)]))
            if show_progress == True:
                self.__progress(i,total, '')

        comparison = kit_tools.ComparedCollocates()
        comparison.read_str('\n'.join(s))
        del s 
        comparison.df.sort_values('DIFFERENCE',ascending=True,inplace=True)
        comparison.df.N = [(i+1)for i in range(0,len(comparison.df))]
        comparison.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return comparison 
    
    def collocations(self,kwic,**kwargs):
        """
        Extracts collocates from kwic
        :param lowercase: Letters are converted to lowercase (default: True)
        :type lowercase: boolean
        :param horizon: horizon (from 1 to 5 - default: 5) 
        :type horizon: int
        :param measure: Statistic measure tscore or mutual information (default: tscore)
        :type measure: str
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Collocations
        :rtype: Collocations
        """
        import re 
        #kwargs
        lowercase = kwargs.get('lowercase',True) 
        horizon = kwargs.get('horizon',5)
        measure = kwargs.get('measure','tscore')
        show_progress= kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # check stat measure
        if measure == 'tscore':
            stat = 1
        elif measure == 'mutual information':
            stat = 2
        # pattern for avoiding punctuation
        ptn = re.compile("^\W+$")
        # wordlist to dic
        tpl = make_wordlist(self.workspace,self.corpus_name,self.language,lowercase)
        tokens = tpl[0]
        freqlist = collections.defaultdict()
        for row in tpl[4]:
            freqlist[row[1]] = row[2]
        # dictionaries and counters
        left_counter = collections.Counter()
        right_counter = collections.Counter()
        words = collections.defaultdict()
        node_freq = 0
        total = len(kwic.df)
        j=0
        for row in kwic.df.itertuples(index=False):
            j+=1
            node_freq+=1
            # get horizons
            if lowercase == True:
                left = list(reversed(row[1].lower().split(' ')[-horizon:]))
                right = row[1].lower().split(' ')[0:horizon]
            else:
                left = list(reversed(row[1].split(' ')[-horizon:]))
                right = row[1].split(' ')[0:horizon]
            # mapping left     
            i = 0
            for word in left:
                if re.search(ptn, word) == None:
                    i+=1
                    left_counter[(i,word)]+=1
                    if word not in words:
                        words[word] = freqlist[word]
            # mapping right
            i=0
            for word in right:
                if re.search(ptn, word) == None:
                    i+=1
                    right_counter[(i,word)]+=1
                    if word not in words:
                        words[word] = freqlist[word]
            if show_progress == True:
                self.__progress(j,total, '')
        del kwic  
        del freqlist 
        # make data table
        s = []
        s.append('\t'.join(['N','WORD','L5','L4','L3','L2','L1','R1','R2','R3','R4','R5','LEFT','RIGHT','TOTAL','ASSOCIATION']))
        i=0
        for word in words:
            total = 0
            total_left = 0
            total_right = 0
            lv = ['0','0','0','0','0','0']
            for hw,freq  in left_counter.most_common():
                if hw[1] == word:
                    lv[hw[0]] = str(freq)
                    total_left+=freq
            rv = ['0','0','0','0','0','0']
            for hw,freq  in right_counter.most_common():
                if hw[1] == word:
                    rv[hw[0]] = str(freq)
                    total_right+=freq
            # calculate stat measure
            total = total_left + total_right
            if stat == 1:
                m = kit_util.tscore(total, words[word], node_freq, tokens, 1)
            elif stat == 2:
                m = kit_util.mutual_information(total, words[word], node_freq, tokens, 1)
            i+=1
            s.append('\t'.join([str(i),word,lv[5],lv[4],lv[3],lv[2],lv[1],rv[1],rv[2],rv[3],rv[4],rv[5],str(total_left),str(total_right),str(total),str(m)]))
        words = None
        left_counter =None
        right_counter = None
        collocations = kit_tools.Collocations()
        collocations.read_str('\n'.join(s))
        del s
        collocations.df.sort_values('ASSOCIATION',ascending=False,inplace=True)
        collocations.df.N = [(i+1)for i in range(0,len(collocations.df))]
        collocations.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return collocations
    
    def clusters(self,word,**kwargs):
        """
        Generates a list of clusters
        :param word: search word
        :type word: str
        :param pos: Part of speech (POS) for each search word 
        :type pos: str or list
        :param size: size of the cluster
        :type size: int
        :param lowercase: Letters are converted to lowercase 
        :type lowercase: boolean
        :param minfreq: minimum frequency of the cluster
        :type minfreq: int
        :param minrange: minimum range frequency 
        :type minrange: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Clusters
        :rtype: Clusters
        """
        # args
        word = word.strip()
        pos = kwargs.get('pos',None)
        size = kwargs.get('size',3)
        lowercase = kwargs.get('lowercase',True)
        min_freq = kwargs.get('minfreq',1)
        min_range = kwargs.get('minrange',1)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        lst_clusters = make_clusters(self.workspace, self.corpus_name, word, pos, lowercase, size)
        clusters = kit_tools.Clusters()
        clusters.df = pd.DataFrame(lst_clusters,columns=['N','CLUSTER','FREQUENCY','RANGE','%'])
        lst_clusters = None
        # sort 
        clusters.df.sort_values('FREQUENCY',ascending=False,inplace=True)
        # filter
        clusters.df=clusters.df[(clusters.df.FREQUENCY >= min_freq)]
        clusters.df=clusters.df[(clusters.df.RANGE >= min_range)] 
        # reset indexes 
        clusters.df['%'] = clusters.df['%'].apply(lambda x: round(x, 2))
        clusters.df.N = [(i+1)for i in range(0,len(clusters.df))]
        clusters.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return clusters 
    
    def ngrams(self,**kwargs):
        """
        Generates a list of n-grams
        :param pos: Part of speech (POS) for each word 
        :type pos: str or list
        :param size: size of the cluster
        :type size: int
        :param lowercase: Letters are converted to lowercase 
        :type lowercase: boolean
        :param minfreq: minimum frequency of the cluster
        :type minfreq: int
        :param min_range: minimum range frequency 
        :type min_range: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Clusters
        :rtype: Clusters
        """
        # args
        pos = kwargs.get('pos',None)
        size = kwargs.get('size',3)
        lowercase = kwargs.get('lowercase',True)
        min_freq = kwargs.get('minfreq',1)
        min_range = kwargs.get('minrange',1)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make ngrams 
        lst_ng = make_ngrams(self.workspace, self.corpus_name, pos, size, lowercase)
        ng = kit_tools.Ngrams()
        ng.df = pd.DataFrame(lst_ng,columns=['N','N-GRAM','FREQUENCY','RANGE','%'])
        lst_ng = None
        # sort 
        ng.df.sort_values('FREQUENCY',ascending=False,inplace=True)
        # filter
        ng.df=ng.df[(ng.df.FREQUENCY >= min_freq)]
        ng.df=ng.df[(ng.df.RANGE >= min_range)] 
        # reset indexes 
        ng.df['%'] = ng.df['%'].apply(lambda x: round(x, 2))
        ng.df.N = [(i+1)for i in range(0,len(ng.df))]
        ng.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return ng
    
    def dispersion(self,node,**kwargs):
        """
        Generates dispersion plots.
        :param node: Search word or phrase (max. 4 words)
        :type node : str
        :param pos: Part of speech (POS) for each search word 
        :type pos: str or list
        :param case_sensitive: Enable regular expression for search word matching
        :type case_sensitive: boolean
        :param regexp: Enable regular expression for search matching
        :type regexp: boolean
        :param limit: Number of concordance lines to return
        :type limit: int
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Dispersion
        :rtype: Dispersion
        """
        # args
        node = node.strip()
        pos = kwargs.get('pos',None)
        case_sensitive = kwargs.get('case_sensitive',False)
        regexp = kwargs.get('regexp',False)
        limit = kwargs.get('limit',None)
        show_progress = kwargs.get('show_progress',False)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # make dispersion
        totals,dpts,lst_disp=make_dispersion(self.workspace, self.corpus_name, node, pos, case_sensitive,regexp)
        disp = kit_tools.Dispersion(output_path=self.output_path)
        disp.df = pd.DataFrame(lst_disp,columns=['N','FILENAME','TOTAL','HITS','S1','S2','S3','S4','S5'])
        disp.dpts = dpts
        disp.total_s1 = totals[0]
        disp.total_s2 = totals[1]
        disp.total_s3 = totals[2]
        disp.total_s4 = totals[3]
        disp.total_s5 = totals[4] 
        lst_disp = None
        # sort 
        disp.df.sort_values('HITS',ascending=False,inplace=True)
        # limit
        if limit != None:
            disp.df = disp.df.head(limit)
        # reset indexes 
        disp.df.N = [(i+1)for i in range(0,len(disp.df))]
        disp.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return disp
    
    def keywords_dispersion(self,keywords,**kwargs):
        """
        Generates dispersion plots from keywords.
        :param limit: Maximum number of keywords to use
        :type limit: int
        :param lowercase: Letters are converted to lowercase 
        :type lowercase: boolean
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        :return: Dispersion
        :rtype: Dispersion
        """
        # args
        lowercase =kwargs.get('lowercase',True)
        limit = kwargs.get('limit',25)
        show_progress = kwargs.get('show_progress',False)
        # deal with args
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # process
        # get keywords
        dict_keywords = dict()
        try:
            i=0
            for row in keywords.df.itertuples(index=False):
                i+=1
                dict_keywords[row[1]] = row[3]
                if i >= limit:
                    break
        except Exception as e:
            print(e) 
        # make dispersion
        totals,dpts,lst_disp=make_keywords_dispersion(self.workspace, self.corpus_name,dict_keywords,lowercase)
        disp = kit_tools.KeywordsDispersion(output_path=self.output_path)
        disp.df = pd.DataFrame(lst_disp,columns=['N','WORD','KEYNESS','HITS','S1','S2','S3','S4','S5'])
        disp.dpts = dpts
        disp.total_s1 = totals[0]
        disp.total_s2 = totals[1]
        disp.total_s3 = totals[2]
        disp.total_s4 = totals[3]
        disp.total_s5 = totals[4] 
        lst_disp = None
        # sort 
        disp.df.sort_values('KEYNESS',ascending=False,inplace=True)
        # limit
        if limit != None:
            disp.df = disp.df.head(limit)
        # reset indexes 
        disp.df.N = [(i+1)for i in range(0,len(disp.df))]
        disp.df.reset_index(drop=True,inplace=True)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return disp 
    
    def combolist(self,**kwargs):
        """
        Creates a wordlist with extended information.
        :param max_pos: Maximum number of POS
        :type max_pos: int
        :param min_freq: Minimum frequency of words
        :type min_freq: int
        :param min_range: Minimum range of words
        :type min_range: int
        :param negative_keyness: Include negative keywords
        :type negative_keyness: boolean
        :param limit: Maximum number of words
        :type limit: int
        :param cat: A custom function to apply to combolist rows and return results to CAT column
        :type cat: function
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        """
        # args
        min_freq = kwargs.get('min_freq',1)
        min_range = kwargs.get('min_range',0)
        negative_keyness = kwargs.get('negative_keyness',True)
        cat = kwargs.get('cat',None)
        limit = kwargs.get('limit',None)
        show_progress = kwargs.get('show_progress',False)
        max_pos = kwargs.get('max_pos',1)
        # time start
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        # wordlist
        if show_progress == True : print('Wordlist...')     
        df = self.wordlist().df.filter(['WORD','FREQUENCY','%']) 
        df.columns= ['WORD','FREQUENCY','F%']
        df=df[(df.FREQUENCY >= min_freq)] 
        # keywords 
        if show_progress == True : print('Keywords...')
        keywords = self.keywords().df.filter(['WORD','KEYNESS'])
        df = pd.merge(df,keywords,on=['WORD','WORD'])
        keywords = None
        if negative_keyness == False:
            df=df[(df.KEYNESS >= 0)] 
        # wfreqinfiles
        if show_progress == True : print('Freqinfiles...')
        wfreqinfiles = self.wfreqinfiles().df.filter(['WORD','RANGE','%'])
        wfreqinfiles.columns=['WORD','RANGE','R%']
        df = pd.merge(df,wfreqinfiles,on=['WORD','WORD'])
        wfreqinfiles = None
        df=df[(df['R%'] >= min_range)] 
        # wtfreq
        if show_progress == True : print('Wtfreq...')
        wtfreq = self.wtfreq()
        d = {}
        if max_pos == 1:
            for row in wtfreq.df.itertuples(index=False):
                if row[1] not in d:
                    d[row[1]]= row[2]
        else:
            for row in wtfreq.df.itertuples(index=False):
                if row[1] not in d:
                    d[row[1]]= row[2]
                else:
                    if len(d[row[1]].split(' ')) < max_pos:
                        d[row[1]] += ' ' + row[2]
        wtfreq=None
        wt = pd.DataFrame(list(d.items()),columns=['WORD','POS'])
        d = None
        df = pd.merge(df,wt,on=['WORD','WORD'])
        wt = None
        # stoplist
        if show_progress == True : print('Stoplist...')
        stoplist = {}
        stoplist_path = self.resource_data_path + 'stoplist_' + self.language + '.tab'
        with open(stoplist_path,'r',encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip())!=0:
                    stoplist[line.strip()]=1
        for row in df.itertuples(index=False):
            if row[0] not in stoplist:
                stoplist[row[0]]=0
        stop = pd.DataFrame(list(stoplist.items()),columns=['WORD','STOPWORD'])
        stoplist = None 
        df = pd.merge(df,stop,on=['WORD','WORD'])
        stop = None
        # lemmas
        if show_progress == True : print('Lemmas...') 
        lemmas_path = self.resource_data_path + '/lemmas_' + self.language + '.tab'
        if os.path.exists(lemmas_path): # if lemmas exist:
            # get lemmas from file
            lemmas = {}
            with open(lemmas_path,'r',encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        f = line.strip().split('\t')
                        if len(f) >= 2:
                            lemmas[f[0].strip()]= f[1].strip()
            # make list of lemmas
            lemma_data = []
            for row in df.itertuples(index=False):
                if row[0] in lemmas:
                    lemma_data.append(lemmas[row[0]])
                else:
                    lemma_data.append('-')
            lemmas=None
            # add to df
            df['LEMMA'] = lemma_data 
            lemma_data = None
        else: # no lemmas available:
            # fill with empty
            lemma_data = ['-']*len(df)
            df['LEMMA'] = lemma_data 
            lemmas=None
            # add to df 
            df['LEMMA'] = lemma_data 
            lemma_data = None
        # CAT
        cat_list = ['-']*len(df)
        df['CAT'] = cat_list
        cat_list = None
        # organize
        df['N'] = [(i+1)for i in range(0,len(df))]
        df = df.filter(['N','WORD','POS','FREQUENCY','F%','KEYNESS','RANGE','R%','STOPWORD','LEMMA','CAT'])
        # limit
        if limit != None:
            df = df.head(limit)
        # make object
        cmblist = kit_tools.Combolist()
        cmblist.df = df 
        df = None
        # cat
        if cat != None:
            try:
                cmblist.cat(cat)
            except Exception as e:
                print(e)
        # time end
        if show_progress == True:
            t1 = time.time()
            total_time = round(t1 - t0,2) 
            print('')
            print('Total time: %s seconds' % total_time)
        return cmblist
    
    # EXPORT CORPUS
    
    def export(self,target_folder,**kwargs):
        """Exports the current corpus to a given destination.
        :param target_folder: Path folder to save the corpus
        :type target_folder: str
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        """
        show_progress = kwargs.get('show_progress',True)
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        if os.path.exists(target_folder) and os.path.isdir(target_folder):
            try:
                subprocess.call(['python',self.__path +'/export_corpus.py', self.workspace,self.corpus_name, target_folder])
            except Exception as e:
                if show_progress == True:
                    print("Kitconc cannot export the corpus.\n")
                    print(e)
            finally:
                if show_progress == True:
                    t1 = time.time()
                    total_time = round(t1 - t0,2) 
                    print('')
                    print('Total time: %s seconds' % total_time)
        else:
            if show_progress == True:
                print('The folder path is not valid.')
                
    # IMPORT CORPUS     
    def add_from_export(self,filename,**kwargs):
        """Imports a corpus to the current workspace.
        :param filename: Path to the corpus file (.zip)
        :type filename: str
        :param show_progress: Prints a progress message if value is True
        :type show_progress: boolean
        """
        show_progress = kwargs.get('show_progress',False)
        if show_progress == True:
            print('Running...')
            t0 = time.time()
        if os.path.exists(filename) and os.path.isfile(filename):
            try:
                import zipfile
                with zipfile.ZipFile(filename, 'r') as zip_ref:
                    zip_ref.extractall(self.workspace)
            except Exception as e:
                if show_progress == True:
                    print('Kitconc cannot import the corpus.\n')
                    print(e)
            finally:
                if show_progress == True:
                    t1 = time.time()
                    total_time = round(t1 - t0,2) 
                    print('')
                    print('Total time: %s seconds' % total_time)
        else:
            if show_progress == True:
                print('The file path is not valid.')
    
    
    # GENERAL FUNCTIONS 
    
    def info(self):
        """Returns statistical information from the corpus (tokens,types,TTR,hapax)
        :return: tuple
        :rtype: tuple
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.info_get()
    
    def tokens(self):
        """Returns the number of tokens.
        :return: int
        :rtype: int
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.tokens_get()
    
    def types(self):
        """Returns the number of types.
        :return: int
        :rtype: int
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.types_get()
    
    def ttr(self):
        """Returns the type/token ratio.
        :return: float
        :rtype: float
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.ttr_get()
    
    def hapax(self):
        """Returns the number of hapax legomena.
        :return: int 
        :rtype: int 
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.hapax_get()
    
    def textfiles(self):
        """Returns the names of textfiles.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.textfiles_get_names()
    
    def fileids(self):
        """Returns the list of file ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.fileids_get()
    
    def words(self,fileids=None):
        """Returns the words from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.words_get(fileids)
    
    def tagged_words(self,fileids=None):
        """Returns the tagged words from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.tagged_words_get(fileids)
    
    def sents(self,fileids=None):
        """Returns sentences from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.sents_get(fileids)
    
    def tagged_sents(self,fileids=None):
        """Returns tagged sentences from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.tagged_sents_get(fileids)
    
    def ndarrays(self,fileids=None):
        """Returns ndarrays from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        # [ 2  2  1  1] = [ word_id tag_id sent_id file_id ]
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.ndarrays_get(fileids)
    
    def ndarrays_filenames(self,fileids=None):
        """Returns ndarrays filenames from textfiles by ids.
        :return: generator
        :rtype: generator
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.ndarrays_filenames_get(fileids) 
    
    def dict_words(self):
        """Returns a dictionary of id and word.
        :return: dict
        :rtype: dict
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.dict_words_get()
    
    def dict_tags(self):
        """Returns a dictionary of id and tag.
        :return: dict
        :rtype: dict
        """
        kd = kit_data.KitData(self.workspace,self.corpus_name,self.language)
        return kd.dict_tags_get()
        
    
   
    
    
        
    
        
    
        
        
        

    
    